#include <gb/gb.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>		// For e.g. abs()

#include <gb/bgb_emu.h>

// Game Boy Competition 2021 entry.
// https://itch.io/jam/gbcompo21 
// T-minus one month and a couple days. 
// Thinking a tunnel rotzoomer, shamelessly inspired by one in "Gejmbaj" by Snorpung & Nordloef. 
	// Theirs uses 40x18 logical pixels in 16-ish shades, which I infer means they're doing 
		// tile compression via scanline shenanigans. 
	// Hence using GBDK's text_scroller.c as a basis. 
	// Notably, this code initially uses an interrupt every scanline, which is undesirable, 
		// since we need all the cycles we can get for some indirection-heavy rendering. 
	// Not exactly Quake, here, but still a lookup gimmick that Snorpung only gets 20-30 Hz out of. 
	// Pros: well-worn demoscene territory, double-buffered (probably), fancy rotation.
	// Cons: extremely low-res by default, speed is not a gimme, kind of a one-trick pony. 
	// The initial goal is STUN Runner. I do love and appreciate that game, 
		// but not so much that I'm afraid to ignore "correct" behavior and do something cool. 
	// But I just about giggled myself to sleep over the idea of rotzoomer Half-Life. 
	// The tunnel gimmick is perspective-correct-ish for tight hallways. 
	// Walls and side passages are possible. Even in STUN Runner, I expect to do cutouts and skyboxes.
	// For now - fixed projection, flying through a straight tunnel. 
// Learning GBDK as I go. Some of this is so-so because it disguises magic. 
	// In particular, some demos use a bitmap, which is not strictly supposed to be possible. 
	// I presume there's a scanline gimmick that swaps BG memory from starting at 0 to starting at 128. 
	// Which is fine, and doesn't mess with sprites, but it can cause weirdness, 
		// and it must have an associated performance cost. It should be less hand-wavy. 
	// Basically, magic is fine, but make please the invocation obvious. 




const uint8_t scanline_offsets_tbl[] = {0, 1, 2, 3, 3, 2, 1, 0, 0, 1, 2, 3, 3, 2, 1, 0};
const uint8_t * scanline_offsets = scanline_offsets_tbl;

#define SCROLL_POS 15
#define SCROLL_POS_PIX_START (SCROLL_POS * 8) - 1
#define SCROLL_POS_PIX_END ((SCROLL_POS + 1) * 8) - 1

uint8_t scroller_x = 0;

const uint8_t scroller_text[] = "This is a text scroller demo for GBDK-2020. You can use ideas, that are "\
"shown in this demo, to make different parallax effects, scrolling of tilemaps which are larger than 32x32 "\
"tiles and TEXT SCROLLERS, of course! Need to write something else to make this text longer than 256 characters. "\
"The quick red fox jumps over the lazy brown dog. 0123456789.          ";

// Some of these are redundant holdovers from the original text_scroller.c code. I forget which. 
//const uint8_t* scroller_next_char = scroller_text;
uint8_t* scroller_vram_addr;
uint16_t base, limit;

uint8_t* vram_pointer; 


// Tube stuff
uint8_t outbyte, angle, screen_distance, distance, input, jitter; 		// Rendering scratch variables 
uint8_t scx, scy; 		// Vertical sync / vsync buffers for SCX/SCY_REG

int8_t gravity; 		// Physics scratch 
int8_t rotation = 11; 		// Game state
int8_t rotation_impulse = 10; 		// Input accumulator
int8_t velocity = 0; 		// Necessary for momentum
int16_t curve_x = 0; 		// Curve offset vectors - goofy sampling gimmick
int16_t curve_y = 0; 
uint8_t tube_color[256]; 		// Might make this multi-dimensional. 
	// E.g. tube_color[2][256] for alternating patterns, 
		// or straight-up using this to encode our level designs. 
	// Some map could be sections 0, 0, 0, 7, 3, and that'd be normal shit followed by cutouts or whatever. 
// Eventually need centripital force variable. Though curve info may suffice. (If all tracks are level.) 
uint8_t controls_locked; 		// Boolean. Over-the-top lock. 

// FPS stuff
uint16_t line_count = 0; 		// This frame 
uint16_t running_average = 1024; 		// Over time - nonzero initial value






/*
// Banged this out in JS console because GBDK has no trig library. 
for( j = 0; j < 1; j++ ) { 
n = 0; d = 0; p = 0; z = 0; t += 1; screen = ""; xmax = 32; ymax = 32;
for( sy = 0; sy < ymax; sy++ ) { screen += "\n"; 
for( sx = 0; sx < xmax; sx++ ) { 
x = ( sx - ( xmax - 1 ) / 2 ); y = ( sy - (ymax-1)/2 ); y *= -1; 
d = Math.hypot( x * 20 / xmax, y * 18 / ymax ); 
z = 128/(d+0.5);
if( x > 0  && y > 0 ) { n = parseInt( Math.atan( Math.abs( y/x ) ) * 360 / ( 2 * Math.PI ) ) } 
else if ( x < 0 && y > 0 ) { n = parseInt( 180 - Math.atan( Math.abs( y/x ) ) * 360 / ( 2 * Math.PI ) ) }
else if ( x < 0 && y < 0 ) { n = parseInt( Math.atan( Math.abs( y/x ) ) * 360 / ( 2 * Math.PI ) ) + 180 }
else if ( x > 0 && y < 0 ) { n = parseInt( 360 - Math.atan( Math.abs( y/x ) ) * 360 / ( 2 * Math.PI ) ) }
p = n % 120  > 120/2; 
p = p ^ ( z % 10 > 5 ); 
screen += parseInt( n * 256 / 360 ) + ", " ; }  } console.log( screen ); }
*/

// These might become un-const so we can modify them on-the-fly. 
// And they might be better off as one table with interleaved elements because consecutive reads are cheaper. 
// For now, clarity is key. 
const uint8_t distance_table[] = 
/* 		// Original take, but with central 254 block -> 128s
{ 14, 14, 15, 16, 17, 18, 19, 20, 20, 21, 21, 20, 20, 19, 18, 17, 16, 15, 14, 14, 
14, 15, 16, 18, 19, 20, 21, 22, 23, 23, 23, 23, 22, 21, 20, 19, 18, 16, 15, 14, 
15, 16, 18, 19, 21, 22, 24, 25, 26, 27, 27, 26, 25, 24, 22, 21, 19, 18, 16, 15, 
16, 17, 19, 21, 23, 25, 27, 29, 31, 32, 32, 31, 29, 27, 25, 23, 21, 19, 17, 16, 
17, 18, 20, 22, 25, 28, 31, 34, 37, 39, 39, 37, 34, 31, 28, 25, 22, 20, 18, 17, 
17, 19, 21, 24, 27, 31, 36, 41, 47, 50, 50, 47, 41, 36, 31, 27, 24, 21, 19, 17, 
18, 20, 22, 25, 29, 34, 41, 50, 61, 70, 70, 61, 50, 41, 34, 29, 25, 22, 20, 18, 
18, 20, 23, 26, 31, 37, 47, 61, 84, 113, 113, 84, 61, 47, 37, 31, 26, 23, 20, 18, 
18, 21, 23, 27, 32, 39, 50, 70, 113, 128, 128, 113, 70, 50, 39, 32, 27, 23, 21, 18, 
18, 21, 23, 27, 32, 39, 50, 70, 113, 128, 128, 113, 70, 50, 39, 32, 27, 23, 21, 18, 
18, 20, 23, 26, 31, 37, 47, 61, 84, 113, 113, 84, 61, 47, 37, 31, 26, 23, 20, 18, 
18, 20, 22, 25, 29, 34, 41, 50, 61, 70, 70, 61, 50, 41, 34, 29, 25, 22, 20, 18, 
17, 19, 21, 24, 27, 31, 36, 41, 47, 50, 50, 47, 41, 36, 31, 27, 24, 21, 19, 17, 
17, 18, 20, 22, 25, 28, 31, 34, 37, 39, 39, 37, 34, 31, 28, 25, 22, 20, 18, 17, 
16, 17, 19, 21, 23, 25, 27, 29, 31, 32, 32, 31, 29, 27, 25, 23, 21, 19, 17, 16, 
15, 16, 18, 19, 21, 22, 24, 25, 26, 27, 27, 26, 25, 24, 22, 21, 19, 18, 16, 15, 
14, 15, 16, 18, 19, 20, 21, 22, 23, 23, 23, 23, 22, 21, 20, 19, 18, 16, 15, 14, 
14, 14, 15, 16, 17, 18, 19, 20, 20, 21, 21, 20, 20, 19, 18, 17, 16, 15, 14, 14 };
*/

/*
		// Z > 1, avoiding extremey distant middle pixels. 
{ 13, 13, 14, 15, 16, 16, 17, 18, 18, 18, 18, 18, 18, 17, 16, 16, 15, 14, 13, 13, 
13, 14, 15, 16, 17, 18, 19, 20, 20, 21, 21, 20, 20, 19, 18, 17, 16, 15, 14, 13, 
14, 15, 16, 17, 18, 20, 21, 22, 23, 23, 23, 23, 22, 21, 20, 18, 17, 16, 15, 14, 
15, 16, 17, 18, 20, 22, 23, 25, 26, 27, 27, 26, 25, 23, 22, 20, 18, 17, 16, 15, 
15, 16, 18, 20, 22, 24, 26, 29, 31, 32, 32, 31, 29, 26, 24, 22, 20, 18, 16, 15, 
16, 17, 19, 21, 23, 26, 30, 33, 37, 39, 39, 37, 33, 30, 26, 23, 21, 19, 17, 16, 
16, 18, 20, 22, 25, 29, 33, 39, 45, 50, 50, 45, 39, 33, 29, 25, 22, 20, 18, 16, 
16, 18, 20, 23, 26, 31, 37, 45, 57, 69, 69, 57, 45, 37, 31, 26, 23, 20, 18, 16, 
17, 18, 21, 23, 27, 32, 39, 50, 69, 105, 105, 69, 50, 39, 32, 27, 23, 21, 18, 17, 
17, 18, 21, 23, 27, 32, 39, 50, 69, 105, 105, 69, 50, 39, 32, 27, 23, 21, 18, 17, 
16, 18, 20, 23, 26, 31, 37, 45, 57, 69, 69, 57, 45, 37, 31, 26, 23, 20, 18, 16, 
16, 18, 20, 22, 25, 29, 33, 39, 45, 50, 50, 45, 39, 33, 29, 25, 22, 20, 18, 16, 
16, 17, 19, 21, 23, 26, 30, 33, 37, 39, 39, 37, 33, 30, 26, 23, 21, 19, 17, 16, 
15, 16, 18, 20, 22, 24, 26, 29, 31, 32, 32, 31, 29, 26, 24, 22, 20, 18, 16, 15, 
15, 16, 17, 18, 20, 22, 23, 25, 26, 27, 27, 26, 25, 23, 22, 20, 18, 17, 16, 15, 
14, 15, 16, 17, 18, 20, 21, 22, 23, 23, 23, 23, 22, 21, 20, 18, 17, 16, 15, 14, 
13, 14, 15, 16, 17, 18, 19, 20, 20, 21, 21, 20, 20, 19, 18, 17, 16, 15, 14, 13, 
13, 13, 14, 15, 16, 16, 17, 18, 18, 18, 18, 18, 18, 17, 16, 16, 15, 14, 13, 13 };
*/

/*
		// Formula with hypot() accounting for xmax and ymas.
{ 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 
10, 10, 11, 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10, 10, 
10, 11, 12, 13, 14, 15, 16, 17, 17, 18, 18, 17, 17, 16, 15, 14, 13, 12, 11, 10, 
11, 12, 13, 14, 15, 16, 18, 19, 20, 21, 21, 20, 19, 18, 16, 15, 14, 13, 12, 11, 
11, 12, 13, 15, 16, 18, 20, 22, 24, 25, 25, 24, 22, 20, 18, 16, 15, 13, 12, 11, 
12, 13, 14, 16, 18, 20, 23, 26, 29, 31, 31, 29, 26, 23, 20, 18, 16, 14, 13, 12, 
12, 13, 15, 17, 19, 22, 26, 31, 37, 41, 41, 37, 31, 26, 22, 19, 17, 15, 13, 12, 
12, 14, 15, 17, 20, 24, 29, 37, 48, 61, 61, 48, 37, 29, 24, 20, 17, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 41, 61, 106, 106, 61, 41, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 41, 61, 106, 106, 61, 41, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 17, 20, 24, 29, 37, 48, 61, 61, 48, 37, 29, 24, 20, 17, 15, 14, 12, 
12, 13, 15, 17, 19, 22, 26, 31, 37, 41, 41, 37, 31, 26, 22, 19, 17, 15, 13, 12, 
12, 13, 14, 16, 18, 20, 23, 26, 29, 31, 31, 29, 26, 23, 20, 18, 16, 14, 13, 12, 
11, 12, 13, 15, 16, 18, 20, 22, 24, 25, 25, 24, 22, 20, 18, 16, 15, 13, 12, 11, 
11, 12, 13, 14, 15, 16, 18, 19, 20, 21, 21, 20, 19, 18, 16, 15, 14, 13, 12, 11, 
10, 11, 12, 13, 14, 15, 16, 17, 17, 18, 18, 17, 17, 16, 15, 14, 13, 12, 11, 10, 
10, 10, 11, 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10, 10, 
9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9 };
*/

/*
		// 32x20 version for 6-scanline-tall tile compression.
{ 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 
9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 
10, 10, 11, 12, 12, 13, 14, 15, 15, 15, 15, 15, 15, 14, 13, 12, 12, 11, 10, 10, 
10, 11, 11, 12, 13, 14, 15, 16, 16, 16, 16, 16, 16, 15, 14, 13, 12, 11, 11, 10, 
10, 11, 12, 13, 14, 15, 16, 17, 17, 18, 18, 17, 17, 16, 15, 14, 13, 12, 11, 10, 
10, 11, 12, 13, 14, 16, 17, 18, 19, 19, 19, 19, 18, 17, 16, 14, 13, 12, 11, 10, 
11, 12, 13, 14, 15, 17, 18, 20, 21, 21, 21, 21, 20, 18, 17, 15, 14, 13, 12, 11, 
11, 12, 13, 14, 16, 18, 19, 21, 23, 24, 24, 23, 21, 19, 18, 16, 14, 13, 12, 11, 
11, 12, 14, 15, 17, 19, 21, 23, 25, 26, 26, 25, 23, 21, 19, 17, 15, 14, 12, 11, 
11, 13, 14, 16, 18, 20, 23, 25, 28, 30, 30, 28, 25, 23, 20, 18, 16, 14, 13, 11, 
12, 13, 14, 16, 18, 21, 24, 28, 32, 35, 35, 32, 28, 24, 21, 18, 16, 14, 13, 12, 
12, 13, 15, 17, 19, 22, 26, 31, 37, 41, 41, 37, 31, 26, 22, 19, 17, 15, 13, 12, 
12, 13, 15, 17, 20, 23, 28, 34, 43, 50, 50, 43, 34, 28, 23, 20, 17, 15, 13, 12, 
12, 14, 15, 17, 20, 24, 29, 38, 50, 64, 64, 50, 38, 29, 24, 20, 17, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 40, 57, 86, 86, 57, 40, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 42, 63, 119, 119, 63, 42, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 42, 63, 119, 119, 63, 42, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 40, 57, 86, 86, 57, 40, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 17, 20, 24, 29, 38, 50, 64, 64, 50, 38, 29, 24, 20, 17, 15, 14, 12, 
12, 13, 15, 17, 20, 23, 28, 34, 43, 50, 50, 43, 34, 28, 23, 20, 17, 15, 13, 12, 
12, 13, 15, 17, 19, 22, 26, 31, 37, 41, 41, 37, 31, 26, 22, 19, 17, 15, 13, 12, 
12, 13, 14, 16, 18, 21, 24, 28, 32, 35, 35, 32, 28, 24, 21, 18, 16, 14, 13, 12, 
11, 13, 14, 16, 18, 20, 23, 25, 28, 30, 30, 28, 25, 23, 20, 18, 16, 14, 13, 11, 
11, 12, 14, 15, 17, 19, 21, 23, 25, 26, 26, 25, 23, 21, 19, 17, 15, 14, 12, 11, 
11, 12, 13, 14, 16, 18, 19, 21, 23, 24, 24, 23, 21, 19, 18, 16, 14, 13, 12, 11, 
11, 12, 13, 14, 15, 17, 18, 20, 21, 21, 21, 21, 20, 18, 17, 15, 14, 13, 12, 11, 
10, 11, 12, 13, 14, 16, 17, 18, 19, 19, 19, 19, 18, 17, 16, 14, 13, 12, 11, 10, 
10, 11, 12, 13, 14, 15, 16, 17, 17, 18, 18, 17, 17, 16, 15, 14, 13, 12, 11, 10, 
10, 11, 11, 12, 13, 14, 15, 16, 16, 16, 16, 16, 16, 15, 14, 13, 12, 11, 11, 10, 
10, 10, 11, 12, 12, 13, 14, 15, 15, 15, 15, 15, 15, 14, 13, 12, 12, 11, 10, 10, 
9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 
9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9 }; 
*/

/*
		// Aspect ratio correction. (Still fucky.)
{ 11, 12, 13, 14, 15, 16, 18, 19, 20, 21, 21, 20, 19, 18, 16, 15, 14, 13, 12, 11, 
11, 12, 13, 14, 16, 17, 19, 20, 22, 22, 22, 22, 20, 19, 17, 16, 14, 13, 12, 11, 
11, 12, 13, 14, 16, 18, 20, 21, 23, 24, 24, 23, 21, 20, 18, 16, 14, 13, 12, 11, 
11, 12, 13, 15, 16, 18, 20, 23, 24, 26, 26, 24, 23, 20, 18, 16, 15, 13, 12, 11, 
11, 12, 14, 15, 17, 19, 21, 24, 26, 27, 27, 26, 24, 21, 19, 17, 15, 14, 12, 11, 
11, 13, 14, 16, 17, 20, 22, 25, 28, 30, 30, 28, 25, 22, 20, 17, 16, 14, 13, 11, 
12, 13, 14, 16, 18, 20, 23, 27, 30, 33, 33, 30, 27, 23, 20, 18, 16, 14, 13, 12, 
12, 13, 14, 16, 18, 21, 25, 29, 33, 36, 36, 33, 29, 25, 21, 18, 16, 14, 13, 12, 
12, 13, 15, 17, 19, 22, 26, 30, 36, 40, 40, 36, 30, 26, 22, 19, 17, 15, 13, 12, 
12, 13, 15, 17, 19, 23, 27, 32, 39, 45, 45, 39, 32, 27, 23, 19, 17, 15, 13, 12, 
12, 13, 15, 17, 20, 23, 28, 34, 43, 51, 51, 43, 34, 28, 23, 20, 17, 15, 13, 12, 
12, 13, 15, 17, 20, 24, 29, 37, 47, 59, 59, 47, 37, 29, 24, 20, 17, 15, 13, 12, 
12, 14, 15, 17, 20, 24, 30, 38, 52, 70, 70, 52, 38, 30, 24, 20, 17, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 40, 57, 84, 84, 57, 40, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 41, 61, 104, 104, 61, 41, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 42, 63, 124, 124, 63, 42, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 42, 63, 124, 124, 63, 42, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 41, 61, 104, 104, 61, 41, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 18, 21, 25, 31, 40, 57, 84, 84, 57, 40, 31, 25, 21, 18, 15, 14, 12, 
12, 14, 15, 17, 20, 24, 30, 38, 52, 70, 70, 52, 38, 30, 24, 20, 17, 15, 14, 12, 
12, 13, 15, 17, 20, 24, 29, 37, 47, 59, 59, 47, 37, 29, 24, 20, 17, 15, 13, 12, 
12, 13, 15, 17, 20, 23, 28, 34, 43, 51, 51, 43, 34, 28, 23, 20, 17, 15, 13, 12, 
12, 13, 15, 17, 19, 23, 27, 32, 39, 45, 45, 39, 32, 27, 23, 19, 17, 15, 13, 12, 
12, 13, 15, 17, 19, 22, 26, 30, 36, 40, 40, 36, 30, 26, 22, 19, 17, 15, 13, 12, 
12, 13, 14, 16, 18, 21, 25, 29, 33, 36, 36, 33, 29, 25, 21, 18, 16, 14, 13, 12, 
12, 13, 14, 16, 18, 20, 23, 27, 30, 33, 33, 30, 27, 23, 20, 18, 16, 14, 13, 12, 
11, 13, 14, 16, 17, 20, 22, 25, 28, 30, 30, 28, 25, 22, 20, 17, 16, 14, 13, 11, 
11, 12, 14, 15, 17, 19, 21, 24, 26, 27, 27, 26, 24, 21, 19, 17, 15, 14, 12, 11, 
11, 12, 13, 15, 16, 18, 20, 23, 24, 26, 26, 24, 23, 20, 18, 16, 15, 13, 12, 11, 
11, 12, 13, 14, 16, 18, 20, 21, 23, 24, 24, 23, 21, 20, 18, 16, 14, 13, 12, 11, 
11, 12, 13, 14, 16, 17, 19, 20, 22, 22, 22, 22, 20, 19, 17, 16, 14, 13, 12, 11, 
11, 12, 13, 14, 15, 16, 18, 19, 20, 21, 21, 20, 19, 18, 16, 15, 14, 13, 12, 11 }; 
*/

		// 32x20 or 32x32 version - wide, so we can pan, and at leas theoretically tall and y-sheared. 
{ 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 
9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 10, 10, 10, 9, 
9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 
10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 
10, 11, 11, 12, 12, 13, 13, 14, 15, 15, 16, 16, 17, 17, 18, 18, 18, 18, 17, 17, 16, 16, 15, 15, 14, 13, 13, 12, 12, 11, 11, 10, 
10, 11, 11, 12, 13, 13, 14, 15, 15, 16, 17, 18, 18, 19, 19, 19, 19, 19, 19, 18, 18, 17, 16, 15, 15, 14, 13, 13, 12, 11, 11, 10, 
11, 11, 12, 12, 13, 14, 15, 15, 16, 17, 18, 19, 20, 21, 21, 21, 21, 21, 21, 20, 19, 18, 17, 16, 15, 15, 14, 13, 12, 12, 11, 11, 
11, 11, 12, 13, 14, 14, 15, 16, 17, 18, 20, 21, 22, 23, 23, 24, 24, 23, 23, 22, 21, 20, 18, 17, 16, 15, 14, 14, 13, 12, 11, 11, 
11, 12, 12, 13, 14, 15, 16, 17, 18, 20, 21, 22, 24, 25, 26, 27, 27, 26, 25, 24, 22, 21, 20, 18, 17, 16, 15, 14, 13, 12, 12, 11, 
11, 12, 13, 14, 14, 15, 17, 18, 19, 21, 23, 25, 26, 28, 29, 30, 30, 29, 28, 26, 25, 23, 21, 19, 18, 17, 15, 14, 14, 13, 12, 11, 
11, 12, 13, 14, 15, 16, 17, 19, 20, 22, 24, 27, 29, 32, 34, 35, 35, 34, 32, 29, 27, 24, 22, 20, 19, 17, 16, 15, 14, 13, 12, 11, 
12, 12, 13, 14, 15, 16, 18, 20, 21, 24, 26, 29, 33, 36, 40, 41, 41, 40, 36, 33, 29, 26, 24, 21, 20, 18, 16, 15, 14, 13, 12, 12, 
12, 13, 13, 14, 16, 17, 18, 20, 22, 25, 28, 32, 37, 42, 47, 51, 51, 47, 42, 37, 32, 28, 25, 22, 20, 18, 17, 16, 14, 13, 13, 12, 
12, 13, 14, 15, 16, 17, 19, 21, 23, 26, 30, 35, 41, 49, 58, 65, 65, 58, 49, 41, 35, 30, 26, 23, 21, 19, 17, 16, 15, 14, 13, 12, 
12, 13, 14, 15, 16, 17, 19, 21, 24, 27, 31, 37, 44, 56, 72, 91, 91, 72, 56, 44, 37, 31, 27, 24, 21, 19, 17, 16, 15, 14, 13, 12, 
12, 13, 14, 15, 16, 18, 19, 21, 24, 27, 32, 38, 47, 61, 86, 139, 139, 86, 61, 47, 38, 32, 27, 24, 21, 19, 18, 16, 15, 14, 13, 12, 
12, 13, 14, 15, 16, 18, 19, 21, 24, 27, 32, 38, 47, 61, 86, 139, 139, 86, 61, 47, 38, 32, 27, 24, 21, 19, 18, 16, 15, 14, 13, 12, 
12, 13, 14, 15, 16, 17, 19, 21, 24, 27, 31, 37, 44, 56, 72, 91, 91, 72, 56, 44, 37, 31, 27, 24, 21, 19, 17, 16, 15, 14, 13, 12, 
12, 13, 14, 15, 16, 17, 19, 21, 23, 26, 30, 35, 41, 49, 58, 65, 65, 58, 49, 41, 35, 30, 26, 23, 21, 19, 17, 16, 15, 14, 13, 12, 
12, 13, 13, 14, 16, 17, 18, 20, 22, 25, 28, 32, 37, 42, 47, 51, 51, 47, 42, 37, 32, 28, 25, 22, 20, 18, 17, 16, 14, 13, 13, 12, 
12, 12, 13, 14, 15, 16, 18, 20, 21, 24, 26, 29, 33, 36, 40, 41, 41, 40, 36, 33, 29, 26, 24, 21, 20, 18, 16, 15, 14, 13, 12, 12, 
11, 12, 13, 14, 15, 16, 17, 19, 20, 22, 24, 27, 29, 32, 34, 35, 35, 34, 32, 29, 27, 24, 22, 20, 19, 17, 16, 15, 14, 13, 12, 11, 
11, 12, 13, 14, 14, 15, 17, 18, 19, 21, 23, 25, 26, 28, 29, 30, 30, 29, 28, 26, 25, 23, 21, 19, 18, 17, 15, 14, 14, 13, 12, 11, 
11, 12, 12, 13, 14, 15, 16, 17, 18, 20, 21, 22, 24, 25, 26, 27, 27, 26, 25, 24, 22, 21, 20, 18, 17, 16, 15, 14, 13, 12, 12, 11, 
11, 11, 12, 13, 14, 14, 15, 16, 17, 18, 20, 21, 22, 23, 23, 24, 24, 23, 23, 22, 21, 20, 18, 17, 16, 15, 14, 14, 13, 12, 11, 11, 
11, 11, 12, 12, 13, 14, 15, 15, 16, 17, 18, 19, 20, 21, 21, 21, 21, 21, 21, 20, 19, 18, 17, 16, 15, 15, 14, 13, 12, 12, 11, 11, 
10, 11, 11, 12, 13, 13, 14, 15, 15, 16, 17, 18, 18, 19, 19, 19, 19, 19, 19, 18, 18, 17, 16, 15, 15, 14, 13, 13, 12, 11, 11, 10, 
10, 11, 11, 12, 12, 13, 13, 14, 15, 15, 16, 16, 17, 17, 18, 18, 18, 18, 17, 17, 16, 16, 15, 15, 14, 13, 13, 12, 12, 11, 11, 10, 
10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 
9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 
9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 10, 10, 10, 9, 
9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9 }; 

const uint8_t angle_table[] = 
/*
{ 98, 96, 93, 90, 86, 83, 79, 75, 71, 66, 61, 56, 51, 47, 44, 40, 36, 34, 32, 29, 
100, 98, 96, 92, 89, 85, 81, 76, 71, 66, 61, 55, 50, 45, 41, 37, 34, 32, 29, 27, 
103, 100, 98, 96, 92, 88, 83, 78, 72, 66, 60, 54, 48, 43, 39, 34, 32, 28, 26, 24, 
105, 104, 101, 98, 96, 91, 86, 81, 74, 67, 59, 52, 46, 40, 35, 32, 28, 25, 22, 21, 
109, 108, 105, 103, 99, 96, 90, 84, 76, 68, 59, 50, 42, 36, 32, 27, 24, 21, 19, 17, 
113, 111, 109, 107, 104, 100, 96, 88, 80, 69, 57, 46, 38, 32, 26, 22, 19, 17, 15, 14, 
117, 115, 114, 112, 110, 106, 102, 96, 85, 71, 55, 41, 32, 24, 20, 17, 14, 12, 11, 9, 
121, 120, 119, 118, 116, 114, 110, 105, 96, 76, 50, 32, 21, 16, 12, 10, 8, 7, 7, 5, 
125, 125, 125, 124, 123, 123, 121, 119, 114, 96, 32, 12, 7, 5, 4, 3, 2, 2, 2, 2, 
130, 130, 130, 130, 131, 132, 133, 135, 140, 160, 224, 242, 247, 249, 251, 251, 252, 253, 253, 253, 
133, 135, 135, 136, 138, 140, 144, 149, 160, 178, 204, 224, 233, 238, 242, 244, 246, 247, 248, 249, 
137, 139, 140, 142, 145, 148, 152, 160, 169, 183, 199, 213, 224, 230, 234, 238, 240, 242, 243, 245, 
142, 143, 145, 147, 150, 154, 160, 166, 174, 185, 197, 208, 216, 224, 228, 232, 235, 237, 239, 241, 
145, 147, 149, 152, 155, 160, 164, 170, 178, 187, 196, 204, 212, 218, 224, 227, 231, 233, 236, 237, 
149, 150, 153, 156, 160, 163, 168, 174, 180, 187, 195, 202, 209, 214, 219, 224, 226, 229, 232, 233, 
152, 154, 156, 160, 162, 167, 171, 176, 182, 188, 194, 200, 206, 211, 216, 220, 224, 226, 228, 231, 
155, 157, 160, 162, 165, 169, 173, 178, 183, 189, 194, 199, 204, 209, 213, 217, 220, 224, 226, 228, 
157, 160, 162, 164, 168, 172, 175, 179, 184, 189, 194, 199, 203, 207, 211, 214, 218, 221, 224, 226 }; 
*/

/*
		// 32x20 version for compressed display:
{ 95, 93, 90, 87, 84, 81, 77, 73, 69, 65, 61, 57, 53, 49, 46, 42, 39, 36, 34, 32, 
96, 94, 91, 88, 86, 82, 78, 74, 70, 66, 61, 56, 52, 48, 44, 41, 38, 35, 32, 30, 
98, 96, 93, 90, 87, 83, 79, 75, 71, 66, 61, 56, 51, 47, 43, 39, 36, 34, 31, 29, 
99, 97, 94, 91, 88, 84, 81, 76, 71, 66, 61, 56, 51, 46, 42, 38, 35, 32, 29, 27, 
100, 98, 96, 93, 90, 86, 81, 77, 71, 66, 61, 55, 49, 45, 40, 36, 33, 30, 28, 26, 
103, 100, 98, 95, 91, 88, 83, 78, 72, 66, 60, 54, 49, 43, 39, 35, 32, 29, 26, 24, 
104, 103, 100, 97, 93, 90, 85, 79, 73, 66, 60, 53, 47, 41, 36, 33, 29, 27, 24, 22, 
106, 104, 102, 99, 96, 92, 87, 81, 74, 67, 59, 52, 45, 39, 34, 31, 27, 24, 22, 20, 
108, 107, 104, 102, 98, 94, 89, 83, 76, 68, 59, 51, 43, 37, 32, 28, 24, 22, 19, 18, 
110, 109, 107, 104, 101, 97, 92, 86, 78, 68, 58, 49, 41, 34, 29, 25, 22, 19, 17, 16, 
113, 111, 110, 108, 104, 100, 96, 89, 80, 69, 57, 46, 37, 31, 26, 22, 19, 17, 15, 13, 
115, 114, 113, 110, 108, 104, 100, 93, 83, 71, 56, 43, 34, 27, 22, 19, 16, 14, 12, 11, 
118, 117, 115, 114, 112, 109, 104, 98, 88, 72, 54, 39, 29, 22, 17, 14, 12, 11, 9, 8, 
120, 120, 119, 118, 116, 113, 110, 104, 94, 76, 51, 32, 22, 17, 13, 10, 9, 7, 7, 6, 
123, 123, 122, 121, 120, 119, 117, 113, 104, 83, 43, 22, 14, 9, 7, 6, 5, 4, 4, 3, 
126, 125, 125, 125, 125, 125, 123, 122, 119, 104, 22, 7, 4, 3, 2, 2, 1, 1, 1, 0, 
128, 129, 129, 129, 130, 130, 131, 132, 135, 150, 232, 247, 250, 251, 253, 253, 253, 253, 253, 254, 
131, 132, 132, 133, 134, 135, 137, 142, 150, 171, 211, 232, 241, 245, 247, 248, 249, 250, 251, 251, 
134, 135, 135, 137, 138, 141, 145, 150, 160, 179, 204, 222, 232, 238, 241, 244, 246, 247, 248, 248, 
136, 137, 139, 140, 142, 145, 150, 157, 167, 182, 200, 216, 226, 232, 237, 240, 242, 243, 245, 246, 
139, 140, 142, 144, 147, 150, 155, 162, 171, 184, 199, 211, 221, 228, 232, 236, 238, 241, 242, 243, 
141, 143, 145, 147, 150, 154, 159, 165, 174, 185, 197, 208, 217, 224, 228, 232, 236, 238, 239, 241, 
144, 145, 147, 150, 153, 157, 162, 169, 177, 186, 196, 206, 214, 220, 225, 229, 232, 235, 237, 238, 
146, 147, 150, 152, 156, 160, 165, 171, 179, 187, 196, 204, 211, 217, 222, 226, 230, 232, 235, 236, 
148, 150, 152, 155, 159, 162, 167, 173, 180, 187, 195, 202, 209, 215, 220, 224, 227, 230, 232, 234, 
150, 152, 155, 157, 161, 164, 169, 175, 181, 188, 194, 201, 207, 213, 218, 221, 225, 228, 231, 232, 
152, 154, 157, 160, 163, 167, 171, 177, 182, 188, 194, 200, 206, 211, 216, 219, 223, 226, 228, 231, 
154, 156, 158, 161, 164, 168, 173, 177, 183, 189, 194, 199, 205, 209, 214, 218, 221, 224, 226, 228, 
155, 157, 160, 163, 166, 170, 174, 179, 184, 189, 194, 199, 204, 209, 212, 216, 219, 222, 225, 227, 
157, 159, 162, 164, 167, 171, 175, 179, 184, 189, 194, 199, 203, 207, 211, 215, 218, 221, 224, 226, 
158, 160, 163, 166, 169, 172, 176, 180, 184, 189, 194, 198, 202, 206, 210, 214, 216, 219, 222, 224, 
160, 162, 164, 167, 170, 174, 177, 181, 185, 189, 193, 197, 201, 205, 209, 212, 215, 218, 221, 223 }; 
*/

/*
		// Aspect ratio correction, because 32x20 is not remotely square:
{ 96, 94, 93, 91, 89, 88, 86, 83, 81, 79, 77, 75, 72, 70, 67, 64, 62, 59, 56, 54, 
96, 96, 93, 92, 91, 88, 87, 85, 83, 81, 78, 76, 73, 70, 67, 64, 62, 59, 56, 54, 
98, 97, 96, 93, 92, 90, 88, 86, 84, 81, 79, 76, 73, 71, 68, 65, 61, 59, 56, 53, 
100, 98, 97, 96, 93, 92, 90, 88, 85, 83, 80, 77, 74, 71, 68, 65, 61, 59, 55, 52, 
101, 100, 98, 97, 96, 93, 91, 89, 87, 84, 81, 78, 75, 72, 68, 65, 61, 58, 54, 51, 
103, 102, 100, 98, 97, 96, 93, 91, 88, 86, 83, 80, 76, 73, 69, 65, 61, 57, 54, 50, 
105, 103, 102, 100, 99, 97, 96, 93, 91, 88, 85, 81, 78, 73, 69, 66, 61, 57, 53, 49, 
107, 105, 104, 103, 101, 100, 98, 96, 93, 90, 86, 83, 79, 75, 71, 66, 61, 56, 51, 47, 
109, 108, 106, 105, 103, 102, 100, 98, 96, 92, 89, 85, 81, 76, 71, 66, 61, 55, 50, 45, 
111, 110, 109, 108, 106, 105, 103, 100, 98, 96, 92, 88, 83, 78, 72, 66, 60, 54, 48, 43, 
113, 113, 111, 110, 109, 108, 105, 104, 101, 98, 96, 91, 86, 81, 74, 67, 59, 52, 46, 40, 
115, 115, 114, 113, 112, 110, 109, 108, 105, 103, 99, 96, 90, 84, 76, 68, 59, 50, 42, 36, 
118, 118, 117, 116, 115, 114, 113, 111, 109, 107, 104, 100, 96, 88, 80, 69, 57, 46, 38, 32, 
120, 120, 120, 119, 118, 118, 117, 115, 114, 112, 110, 106, 102, 96, 85, 71, 55, 41, 32, 24, 
123, 123, 123, 123, 122, 121, 121, 120, 119, 118, 116, 114, 110, 105, 96, 76, 50, 32, 21, 16, 
126, 126, 125, 125, 125, 125, 125, 125, 125, 124, 123, 123, 121, 119, 114, 96, 32, 12, 7, 5, 
128, 128, 129, 129, 129, 129, 130, 130, 130, 130, 131, 132, 133, 135, 140, 160, 224, 242, 247, 249, 
131, 131, 132, 132, 132, 133, 133, 135, 135, 136, 138, 140, 144, 149, 160, 178, 204, 224, 233, 238, 
134, 134, 135, 135, 136, 137, 137, 139, 140, 142, 145, 148, 152, 160, 169, 183, 199, 213, 224, 230, 
136, 137, 137, 138, 139, 140, 142, 143, 145, 147, 150, 154, 160, 166, 174, 185, 197, 208, 216, 224, 
139, 140, 140, 141, 142, 144, 145, 147, 149, 152, 155, 160, 164, 170, 178, 187, 196, 204, 212, 218, 
141, 142, 143, 144, 145, 147, 149, 150, 153, 156, 160, 163, 168, 174, 180, 187, 195, 202, 209, 214, 
143, 145, 145, 147, 148, 150, 152, 154, 156, 160, 162, 167, 171, 176, 182, 188, 194, 200, 206, 211, 
145, 147, 148, 149, 151, 152, 155, 157, 160, 162, 165, 169, 173, 178, 183, 189, 194, 199, 204, 209, 
147, 149, 150, 152, 153, 155, 157, 160, 162, 164, 168, 172, 175, 179, 184, 189, 194, 199, 203, 207, 
150, 151, 152, 154, 155, 157, 160, 162, 164, 167, 169, 173, 177, 181, 185, 189, 194, 197, 201, 206, 
152, 152, 154, 156, 157, 160, 161, 164, 166, 169, 172, 174, 178, 182, 185, 189, 193, 197, 201, 204, 
153, 155, 156, 157, 160, 161, 163, 165, 167, 170, 173, 176, 179, 182, 186, 189, 193, 196, 200, 203, 
155, 156, 157, 160, 161, 162, 164, 167, 169, 172, 174, 177, 180, 183, 187, 189, 193, 196, 199, 202, 
157, 157, 160, 161, 162, 164, 166, 168, 170, 173, 175, 178, 181, 184, 187, 189, 193, 196, 199, 201, 
158, 160, 161, 162, 164, 166, 167, 169, 172, 174, 177, 179, 182, 184, 187, 190, 192, 195, 198, 201, 
160, 160, 162, 164, 165, 167, 169, 171, 173, 175, 177, 179, 182, 184, 187, 190, 192, 195, 198, 200 }; 
*/

		// 32x32 or possibly 32x20 version for panning and/or y-shearing. 
{ 96, 94, 93, 91, 89, 88, 86, 83, 81, 79, 77, 75, 72, 70, 67, 64, 62, 59, 56, 54, 51, 49, 47, 45, 43, 41, 39, 37, 36, 34, 32, 32, 
96, 96, 93, 92, 91, 88, 87, 85, 83, 81, 78, 76, 73, 70, 67, 64, 62, 59, 56, 54, 51, 49, 46, 44, 41, 39, 38, 36, 34, 33, 32, 30, 
98, 97, 96, 93, 92, 90, 88, 86, 84, 81, 79, 76, 73, 71, 68, 65, 61, 59, 56, 53, 50, 47, 45, 42, 40, 38, 36, 34, 33, 32, 29, 29, 
100, 98, 97, 96, 93, 92, 90, 88, 85, 83, 80, 77, 74, 71, 68, 65, 61, 59, 55, 52, 49, 46, 44, 41, 39, 36, 34, 33, 32, 29, 28, 27, 
101, 100, 98, 97, 96, 93, 91, 89, 87, 84, 81, 78, 75, 72, 68, 65, 61, 58, 54, 51, 48, 45, 42, 39, 37, 35, 33, 32, 29, 28, 27, 25, 
103, 102, 100, 98, 97, 96, 93, 91, 88, 86, 83, 80, 76, 73, 69, 65, 61, 57, 54, 50, 46, 44, 41, 38, 36, 33, 32, 29, 28, 26, 24, 24, 
105, 103, 102, 100, 99, 97, 96, 93, 91, 88, 85, 81, 78, 73, 69, 66, 61, 57, 53, 49, 45, 41, 39, 36, 34, 32, 29, 27, 26, 24, 23, 22, 
107, 105, 104, 103, 101, 100, 98, 96, 93, 90, 86, 83, 79, 75, 71, 66, 61, 56, 51, 47, 44, 40, 36, 34, 32, 29, 27, 25, 24, 22, 21, 19, 
109, 108, 106, 105, 103, 102, 100, 98, 96, 92, 89, 85, 81, 76, 71, 66, 61, 55, 50, 45, 41, 37, 34, 32, 29, 27, 24, 23, 21, 20, 19, 17, 
111, 110, 109, 108, 106, 105, 103, 100, 98, 96, 92, 88, 83, 78, 72, 66, 60, 54, 48, 43, 39, 34, 32, 28, 26, 24, 22, 20, 19, 17, 17, 15, 
113, 113, 111, 110, 109, 108, 105, 104, 101, 98, 96, 91, 86, 81, 74, 67, 59, 52, 46, 40, 35, 32, 28, 25, 22, 21, 19, 17, 16, 15, 14, 13, 
115, 115, 114, 113, 112, 110, 109, 108, 105, 103, 99, 96, 90, 84, 76, 68, 59, 50, 42, 36, 32, 27, 24, 21, 19, 17, 16, 14, 13, 12, 12, 11, 
118, 118, 117, 116, 115, 114, 113, 111, 109, 107, 104, 100, 96, 88, 80, 69, 57, 46, 38, 32, 26, 22, 19, 17, 15, 14, 12, 11, 10, 9, 9, 8, 
120, 120, 120, 119, 118, 118, 117, 115, 114, 112, 110, 106, 102, 96, 85, 71, 55, 41, 32, 24, 20, 17, 14, 12, 11, 9, 9, 8, 7, 7, 6, 6, 
123, 123, 123, 123, 122, 121, 121, 120, 119, 118, 116, 114, 110, 105, 96, 76, 50, 32, 21, 16, 12, 10, 8, 7, 7, 5, 5, 4, 4, 4, 3, 3, 
126, 126, 125, 125, 125, 125, 125, 125, 125, 124, 123, 123, 121, 119, 114, 96, 32, 12, 7, 5, 4, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 
128, 128, 129, 129, 129, 129, 130, 130, 130, 130, 131, 132, 133, 135, 140, 160, 224, 242, 247, 249, 251, 251, 252, 253, 253, 253, 253, 253, 253, 253, 254, 254, 
131, 131, 132, 132, 132, 133, 133, 135, 135, 136, 138, 140, 144, 149, 160, 178, 204, 224, 233, 238, 242, 244, 246, 247, 248, 249, 249, 250, 251, 251, 251, 251, 
134, 134, 135, 135, 136, 137, 137, 139, 140, 142, 145, 148, 152, 160, 169, 183, 199, 213, 224, 230, 234, 238, 240, 242, 243, 245, 246, 246, 247, 248, 248, 248, 
136, 137, 137, 138, 139, 140, 142, 143, 145, 147, 150, 154, 160, 166, 174, 185, 197, 208, 216, 224, 228, 232, 235, 237, 239, 241, 242, 243, 244, 245, 246, 246, 
139, 140, 140, 141, 142, 144, 145, 147, 149, 152, 155, 160, 164, 170, 178, 187, 196, 204, 212, 218, 224, 227, 231, 233, 236, 237, 238, 240, 241, 242, 243, 243, 
141, 142, 143, 144, 145, 147, 149, 150, 153, 156, 160, 163, 168, 174, 180, 187, 195, 202, 209, 214, 219, 224, 226, 229, 232, 233, 236, 237, 238, 239, 241, 241, 
143, 145, 145, 147, 148, 150, 152, 154, 156, 160, 162, 167, 171, 176, 182, 188, 194, 200, 206, 211, 216, 220, 224, 226, 228, 231, 233, 234, 236, 237, 238, 239, 
145, 147, 148, 149, 151, 152, 155, 157, 160, 162, 165, 169, 173, 178, 183, 189, 194, 199, 204, 209, 213, 217, 220, 224, 226, 228, 230, 231, 233, 234, 236, 237, 
147, 149, 150, 152, 153, 155, 157, 160, 162, 164, 168, 172, 175, 179, 184, 189, 194, 199, 203, 207, 211, 214, 218, 221, 224, 226, 228, 229, 231, 232, 233, 235, 
150, 151, 152, 154, 155, 157, 160, 162, 164, 167, 169, 173, 177, 181, 185, 189, 194, 197, 201, 206, 209, 213, 216, 219, 221, 224, 225, 227, 228, 230, 231, 233, 
152, 152, 154, 156, 157, 160, 161, 164, 166, 169, 172, 174, 178, 182, 185, 189, 193, 197, 201, 204, 208, 211, 214, 216, 219, 221, 224, 225, 226, 228, 230, 231, 
153, 155, 156, 157, 160, 161, 163, 165, 167, 170, 173, 176, 179, 182, 186, 189, 193, 196, 200, 203, 206, 209, 212, 215, 217, 219, 221, 224, 225, 226, 228, 229, 
155, 156, 157, 160, 161, 162, 164, 167, 169, 172, 174, 177, 180, 183, 187, 189, 193, 196, 199, 202, 205, 208, 211, 213, 216, 218, 220, 221, 224, 225, 226, 228, 
157, 157, 160, 161, 162, 164, 166, 168, 170, 173, 175, 178, 181, 184, 187, 189, 193, 196, 199, 201, 204, 207, 209, 212, 214, 216, 218, 220, 221, 224, 225, 226, 
158, 160, 161, 162, 164, 166, 167, 169, 172, 174, 177, 179, 182, 184, 187, 190, 192, 195, 198, 201, 204, 206, 209, 211, 213, 215, 216, 219, 220, 221, 224, 224, 
160, 160, 162, 164, 165, 167, 169, 171, 173, 175, 177, 179, 182, 184, 187, 190, 192, 195, 198, 200, 203, 205, 207, 209, 211, 214, 216, 217, 219, 221, 222, 224 }; 

// Trig shit. 
// Sine / cosine table, using the obvious gimmick of a 90-degree offset to overlap them. 
/* JS: 
JSON.stringify( 
	new Array( 256 * 1.5 )
		.fill(0).map( (v,i,a) => Math.floor( 
			127 * Math.cos( Math.PI * i / 256.0 ) 
		) ) 
).split(',').join(', ')
*/
// Not thrilled about the lack of zeroes. 
int8_t cos[] = 
{ 127, 126, 126, 126, 126, 126, 126, 126, 126, 126, 126, 125, 125, 125, 125, 124, 124, 124, 123, 123, 123, 122, 122, 121, 121, 121, 120, 120, 119, 119, 118, 117, 117, 116, 116, 115, 114, 114, 113, 112, 112, 111, 110, 109, 108, 108, 107, 106, 105, 104, 103, 102, 102, 101, 100, 99, 98, 97, 96, 95, 94, 93, 91, 90, 89, 88, 87, 86, 85, 84, 82, 81, 80, 79, 78, 76, 75, 74, 73, 71, 70, 69, 67, 66, 65, 63, 62, 61, 59, 58, 57, 55, 54, 52, 51, 50, 48, 47, 45, 44, 42, 41, 39, 38, 36, 35, 33, 32, 30, 29, 27, 26, 24, 23, 21, 20, 18, 17, 15, 13, 12, 10, 9, 7, 6, 4, 3, 1, 0, -2, -4, -5, -7, -8, -10, -11, -13, -14, -16, -18, -19, -21, -22, -24, -25, -27, -28, -30, -31, -33, -34, -36, -37, -39, -40, -42, -43, -45, -46, -48, -49, -51, -52, -53, -55, -56, -58, -59, -60, -62, -63, -64, -66, -67, -68, -70, -71, -72, -74, -75, -76, -77, -79, -80, -81, -82, -83, -85, -86, -87, -88, -89, -90, -91, -92, -94, -95, -96, -97, -98, -99, -100, -101, -102, -103, -103, -104, -105, -106, -107, -108, -109, -109, -110, -111, -112, -113, -113, -114, -115, -115, -116, -117, -117, -118, -118, -119, -120, -120, -121, -121, -122, -122, -122, -123, -123, -124, -124, -124, -125, -125, -125, -126, -126, -126, -126, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -127, -126, -126, -126, -126, -125, -125, -125, -124, -124, -124, -123, -123, -122, -122, -122, -121, -121, -120, -120, -119, -118, -118, -117, -117, -116, -115, -115, -114, -113, -113, -112, -111, -110, -109, -109, -108, -107, -106, -105, -104, -103, -103, -102, -101, -100, -99, -98, -97, -96, -95, -94, -92, -91, -90, -89, -88, -87, -86, -85, -83, -82, -81, -80, -79, -77, -76, -75, -74, -72, -71, -70, -68, -67, -66, -64, -63, -62, -60, -59, -58, -56, -55, -53, -52, -51, -49, -48, -46, -45, -43, -42, -40, -39, -37, -36, -34, -33, -31, -30, -28, -27, -25, -24, -22, -21, -19, -18, -16, -14, -13, -11, -10, -8, -7, -5, -4, -2 }; 

int8_t* sin = &cos[ 128 ]; 		// ... how the fuck is this a 7? 
	// No, how the fuck does cos(90) go 7, 6, 4, 3, 1... 7... -1, -3, -4? What the fuck, Javascript? 
	// FOR THE LOVE OF GOD. With out parseInt - the value is 7.776507174585693e-15. 
	// How does a language ENTIRELY BASED AROUND DOUBLE FLOATS not properly handle floating point conversions?!
	// What drunken ape cast the result of 127*Math.cos() to a fucking string? Where does it even happen? 
	// Is parseInt() intended to only take strings? Jeeesus. I've been using that for years. 
	// Well okey-fuckin-dokey then. That is a bit more like it. Still doesn't make it to zero at the end, but I guess it wouldn't. 

// None of that matters because using trig for this turns out to suck. 
// Just make a table where the lower side facets push you down, and do a minimum-velocity check to go over the top. 
// Or not even that. It's one division.

// Need a 32x32 lookup table for curves. 
	// For starters, actual hypoteneuse, versus Manhattan distance. 
	// But maybe something that blends to zero on the edges? 
		// Less of a concern now because we're way off-center for a 32x32 logical pattern. 
		// But tracking (TBD) may become an issue. Or we just duplicate those lookups. Duh. 

/* 
I fucking hate dealing with bitmasks in hex, but binary is so verbose. 

0000 = 0
0001 = 1
0010 = 2
0011 = 3

0100 = 4
0101 = 5
0110 = 6
0111 = 7

1000 = 8
1001 = 9
1010 = A
1011 = B

1100 = C
1101 = D
1110 = E
1111 = F

*/

const unsigned char gradient_data[] = { 

	/* 0xFF to 0x00, presumably white to black */
	/* 0b1010 = A, 0b0101 = 5 */
	// Black. Black / dark. Dark. Dark / light. Light. Light / white. White. 
	// Currently have 13 colors via simple dithering in 4-pixel patterns. Need a few more for easy power-of-two 16. 
	// Could just stuff some duplicate darks in there, for later? Dark is probably where we want more detail. 

	// BB/BB
	0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF, 		// Black: 1/1

	// BB/BD
	0xFF, 0xFF, 0x77, 0xFF, 0xFF, 0xFF, 0xDD, 0xFF, 0xFF, 0xFF, 0x77, 0xFF, 0xFF, 0xFF, 0xDD, 0xFF, 		// Pips
	0xFF, 0xFF, 0xAA, 0xFF, 0xFF, 0xFF, 0xAA, 0xFF, 0xFF, 0xFF, 0xAA, 0xFF, 0xFF, 0xFF, 0xAA, 0xFF, 		// BB/BB

	// BD/BD
	// B is 1/1, D is 0/1
	0b01010101, 0xFF, 
	0b11101110, 0xFF, 
	0b01010101, 0xFF, 
	0b10111011, 0xFF,

	0b01010101, 0xFF, 
	0b11101110, 0xFF,
	0b01010101, 0xFF, 
	0b10111011, 0xFF,
	0x55, 0xFF, 0xAA, 0xFF, 0x55, 0xFF, 0xAA, 0xFF, 0x55, 0xFF, 0xAA, 0xFF, 0x55, 0xFF, 0xAA, 0xFF, 		// BD/BD 

	// BD/DD
/*
0x15, 0xFF, 
0x8A, 0xFF, 
0x51, 0xFF, 
0xA8, 0xFF, 

0x15, 0xFF, 
0x8A, 0xFF, 
0x51, 0xFF, 
0xA8, 0xFF, 
*/
	0x55, 0xFF, 0x00,0xFF, 0x55, 0xFF, 0x00,0xFF, 0x55, 0xFF, 0x00,0xFF, 0x55, 0xFF, 0x00,0xFF, 		// BD/DD 

	// DDDD DDDD / BDDD BDDD / DDDD DDDD / DDBD DDBD 
	// Or should it be DDDB / DBDD? Argh. 
	0x00, 0xFF, 
	0b00010001, 0xFF, 
	0x00, 0xFF, 
	0b01000100, 0xFF, 

	0x00, 0xFF, 
	0b00010001, 0xFF, 
	0x00, 0xFF, 
	0b01000100, 0xFF,  

	// DD/DD
	0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF, 		// Dark: 0/1

	// DD/DL
	0x00,0xFF, 0x55, 0xAA,0x00,0xFF, 0x55, 0xAA,0x00,0xFF, 0x55, 0xAA,0x00,0xFF, 0x55, 0xAA,

	// DL/DL
	0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA,  	// 0/1 + 1/0 

	// DL/LL
	0xAA, 0x55, 0xFF, 0x00, 0xAA, 0x55, 0xFF, 0x00, 0xAA, 0x55, 0xFF, 0x00, 0xAA, 0x55, 0xFF, 0x00, 

	// LL/LL
	0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00,0xFF,0x00, 		// Light: 1/0

	// LL/LW
	0xFF, 0x00, 0xAA, 0x00, 0xFF, 0x00, 0xAA, 0x00, 0xFF, 0x00, 0xAA, 0x00, 0xFF, 0x00, 0xAA, 0x00, 

	// LW/LW
	0x55, 0x00, 0xAA, 0x00, 0x55, 0x00, 0xAA, 0x00, 0x55, 0x00, 0xAA, 0x00, 0x55, 0x00, 0xAA, 0x00, 	// 1/0 + 0/0

	// LW/WW
	0x55, 0x00, 0x00, 0x00, 0x55, 0x00, 0x00, 0x00, 0x55, 0x00, 0x00, 0x00, 0x55, 0x00, 0x00, 0x00, 

	// WW/WW
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00, 		// White: 0/0
	// This needs to not go white-white.
	// Good answers? Dunno. Just mix with black for now. 

//	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 		// White again

	// BW/BW
	0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA, 0x55, 0x55, 0xAA, 0xAA 
		// Black+white

};





// Interrupt functions

uint8_t sy = 0; 

void half_tile_isr() { 
	// Okay: SCY_REG is a -global- scroll register. It is not "which scanline are we on, right now."
	// It changes which scanline is at the logical top of the screen. 
	// So we can't just sy*8 to go, "next scanline, start drawing from tile row sy."
	// This kinda-sorta makes sense, but it is a tremendous pain in the ass for comprehension, 
		// when mixed with LYC_REG needing a strict objective "where are we NOW?" number. 
	// TL;DR LYC to sy*i and SCY to k where i+k = 8. 
	LYC_REG = sy * 5;    
	SCY_REG += 3; 
//	SCX_REG += 1; 		// Tilting for a turn effect? Eh. 
	sy++; 

	// Debug, making row heights crystal clear:
//	SCX_REG += 128; 

	// So why the fuck is the top row still 7 pixels? Whatever. 
	// I'd have to switch to the window to get 4px-tall tiles. 
	// Or: naive 4px rows, plus a narrow hud at the bottom. linear map, maybe. 
	// Or: 5px rows, and we only waste one scanline. But the dithering might mismatch. (Yeah, it does.) 
	// Dithering could be handled by alternating SCY_REG increments. 

	line_count += 4; 		// FPS guff 
}

void my_vblank_isr() { 
//	LYC_REG = 1; 
//	SCY_REG = 0; 		// Stop doing this if we're not fucking with tilemap compression. 
//	SCX_REG = -16; 
//	sy = 1; 

	// Vsync frame swap, if applicable:
	// SCX_REG = scx; 
	// SCY_REG = scy; 

	line_count += 144; 		// FPS tracking. Not 154 because the scanline version doesn't count vblank. 
	line_count += 10; 		// Account for all scanlines. Probably. Might be missing a few anyway. 
}


uint8_t x, y, xplus; 
int8_t bend_x = 0; 
int8_t bend_y = 0; 
void draw_tile();







// Main functionality 

void main() {
//	const uint8_t* angle_table_split = &angle_table[16*20]; 
//	const uint8_t* distance_table_split = &distance_table[16*20]; 
	printf( "Loading..." ); // Need a printf for GBDK to include text characters. 

	// Load tile data to CHR-RAM
	set_bkg_data(0x00, 0x08, gradient_data);
	set_bkg_data(128, 128, gradient_data);  

	// Fill tube_color because division by six is weirdly expensive. 
	// "Anything that only happens once is free."
	// Weirdly difficult to get a loop of 256. 
	const uint8_t sides[] = { 2, 0, 1, 2, 3, 4, 2, 0 }; 		// 8 instead of 6 because modulo is acting weird. 
	for( uint8_t n = 255; n > 0; n-- ) { 		// Divide into sixths. 
		const uint8_t side = n / 43 + 1; 	// To simplify dealing with non-uniform gradients. 
		tube_color[n] = sides[ side ]; 		// 0..5. 6*43 is 258. 

		// Faux subsampling? 0..5 x2 -> 0..10, and use an in-between color right near the boundary. 
		// "Anti-aliasing." Three colors per side, or just two? Three fits but may be overkill. 
		// Keep dead code for now; we're gonna mix things up during play. 
		tube_color[n] *= 2; 
//		if( n % 43 > 40 ) { tube_color[n]++; } 
//		if( n % 43 < 2 ) { tube_color[n]--; } 
		// Wow, that's nice. 
//		if( n % 43 > 40 ) { tube_color[n] += sides[ ( side + 1 ) % 6 ]; tube_color[n] /= 2; } 
//		if( n % 43 < 2 ) { tube_color[n] += sides[ ( side - 1 ) % 6 ]; tube_color[n] /= 2; } 
		if( n % 43 > 40 ) { 		// Why is anti-aliasing uneven? This is only 41, 42, but the other's only 0, 1. 
											// The bottom facet's diagonals are noticeably askew. 
											// ... maybe just because it's 4->2 instead of 3? Ehh, not really.  
			if( sides[ side + 1 ] > sides[ side ] ) { tube_color[n]++; } 
			else { tube_color[n]--; } 
		}
		else if( n % 43 < 2 ) { 
			if( sides[ side - 1 ] > sides[ side ] ) { tube_color[n]++; } 
			else { tube_color[n]--; } 
		}

		// This setup loop should also handle most of the tube-color stuff. 
		// So e.g. add 3 to every thing for a fairly naive sawtooth gradient that looks about right. 
		// But ideally I'd like to make the bottom white(ish), the upper-right dark, 
			// and the lower-right about as dim as the upper-left. 
		tube_color[n] += 3; 

		tube_color[n] += 128; 		// Tiles in the middle. (Letters at the front.) 
			// I could use a bit to indicate boundaries.
			// Really I could use several, we're only encoding 16 shades so far. 
			// Masking is cheap and it'd let us substitute rotation-aware diagonals. 
	}
	// Then set tube_color[0] if it's not supposed to be black. 
	tube_color[0] = tube_color[1]; 		// Easy hack. 

	// Compression to 4-scanline-high tiles has tolerable performance impact. 
	// It's rendering all those tiles that kills us. 
	// We might go for diagonal tiles and skip scanline shenanigans altogether. 
	// In which case: we do still need VBL_IFLAG. (Or is that just for wait_vbl_done?) 
	CRITICAL {
		STAT_REG |= 0b01000000; LYC_REG = 0;
//		add_LCD( half_tile_isr );
		add_VBL( my_vblank_isr ); 		// Two hard problems. 
//		set_interrupts( VBL_IFLAG | LCD_IFLAG );
		set_interrupts( VBL_IFLAG );
	}

	// FPS / HUD - using window 
	init_win( 6 ); 		// All white, probably. 
	move_win( 7, 128 ); 		// Coordinates start one tile/sprite offscreen. (144-16) throws a warning. Shrug. 
	SHOW_WIN; 		// Equivalent to LCDC_REG|=0x20U

	// Initial game state
	curve_x = 0; 		// 512 is a scientific wild-ass guess. 
	curve_y = 0; 
		// Uncertain fixed-point format. 		
		// But division's going to be with a number from 0..128, so...
		// No, hang on, the result needs to be in the range of like 0..10. Or less. 
		// Large results are probably fine because we can shift down and still benefit from accuracy. 
		// Nope, went with multiplicative solution. Small numbers only, for now. 

		int8_t pan_x = 0; 		// Debug 

	// Make it work - make it fast - make it pretty. In that order. 
	while ( 1 ) {
		// Book-keeping
		scroller_x++;

		// Input
		input = joypad(); 

		// Physics
		// Might just rename rotation_impulse, etc., to velocity, acceleration, and snap. 
		// Fuck it, just make the lower sides slide your ass down.
		if( ! ( input && ( J_LEFT || J_RIGHT ) ) ) { rotation_impulse = 0; } 		// Added force drops to zero instantly. 
		else {
			if( input & J_LEFT ) { 
				rotation_impulse -= 1; 
				if( rotation_impulse < -5 ) { rotation_impulse = -3; } 
			} 
			if( input & J_RIGHT ) { 
				rotation_impulse += 1; 			
				if( rotation_impulse > 5 ) { rotation_impulse = 3; } 
			} 
		}
		// Bit squirrelly at 10 FPS, but it'll do for now. 
		velocity += rotation_impulse; 


		const uint8_t side = ( (int8_t) rotation ) / 43 + 1; 
			// Uh, how does the singed / unsigned thing work here? 
			// I'd rather not need to shift to 16-bit to get that right. 
			// If this works out like [], the top is sides 0-1-2, and the bottom is 3-4-5, with 4 being level. 
		if( side == 3 ) { velocity += 1; } 
		if( side == 5 ) { velocity -= 1; } 

		// Double fuck it. Conditionals. 
		// And lock controls when going over the top. 
			// Going over? Set boolean. Boolean checked? Check for when you're back on the bottom half. 
		// How fast is over the top? 25-ish is the limit for jostling back and forth, trying to get a run-up. 27, tops. 
		// Run-up from the opposite edge of the bottom facet: 17-ish. 
		// Starting from a astandstill at the edge of this facet? 7-ish. 
		// Oh god I'm not sure you should be able to go over in a curve. 
			// Like, if forces push you away, toward the outside, you can build speed toward the midpoint cutoff -
				// but realistically it would keep you there.
			// For smooth "round" gravity, this was going to be a non-issue, because you'd just change the zero point. 

		// Most of this should be DRY-er. It helps keep things consistent. 
		if( rotation < -64 && velocity < 0 ) { 
			if( abs( velocity ) > 25 ) { 
				controls_locked = 1; 		// No proper bools. 
			} 
			else { 
//				velocity = 0; 
				velocity -= velocity / 2; 
				velocity -= -2; 
			}
		} 
		if( rotation > 64 && velocity > 0 ) { 
			if( abs( velocity ) > 25 ) { 
				controls_locked = 1; 		// No proper bools. 
			} 
			else {
//				velocity = 0; 
				velocity -= velocity / 2; 
				velocity -= 2; 
			}
		} 
		if( rotation < -21 ) { velocity += 2; } 
		if( rotation > 21 ) { velocity -= 2; } 

		// Sudden loss of speed when you go over the top?
		// Once again - works great in one direction, fucked up in the other direction. Fuck's sake. 
		if( controls_locked && rotation < -(127-21) && velocity > 0 ) { velocity = velocity / 2; } 
		if( controls_locked && rotation > (127-21) && velocity < 0 ) { velocity = velocity / 2; } 

		// Unlock condition is: in bottom half, headed toward zero. 
		if( controls_locked && rotation > -64 && velocity > 0 ) { 
			velocity -= velocity / 2; 
			controls_locked = 0; 
		} 
		if( controls_locked && rotation < 64 && velocity > 0 ) { 
			velocity -= velocity / 2; 
			controls_locked = 0; 
		} 

//		velocity = velocity > 30 ? 30 : velocity; 
//		velocity = velocity < -30 ? -30 : velocity; 
			// Debug-ish limit on speed, to make endless looping at least happen less on its own, 
				// and also so you can't hold one direction while going over and spin like a moron. 
			// Oh goody, and we're back to being asymmetrical. Fuck you. 

//		rotation += velocity; 

		// Jank-ass basic controls for testing graphics: 
		if( input & J_LEFT ) { rotation -= 4; } 
		if( input & J_RIGHT ) { rotation += 4; } 

		// Friction AFTER movement, so I can add 1 and not get instantly zeroed out. 
		// You can still get stuck in a pendulum loop after going over the top. 
			// Oh, probably just because we never unlock controls. (Which brings back "ramp" gravity.) 
		if( ! controls_locked ) { 
			if( velocity > 0 ) { velocity--; } 
			if( velocity < 0 ) { velocity++; } 
		}



		// Curve test
		if( input & J_UP ) { curve_x += 1; } 
		if( input & J_DOWN ) { curve_x -= 1; } 
		if( input & J_B ) { pan_x += 5; } 
		if( input & J_A ) { pan_x -= 5; } 

		// Y-shear test
/*
		if( input & J_UP ) { ystart--; } 
		if( input & J_DOWN ) { ystart++; } 
		if( input & J_B ) { xstart--; } 
		if( input & J_A ) { xstart++; } 
*/
		#define WIDTH 32
		#define HEIGHT 32		// Possibly 18 or 20


		// Screen placement
		// Use fine hardware scrolling, you ding-dong. 
		const int8_t half_rotation = rotation / 2; 

		// "Evelyn" keeps barking about these.
		// Possibly because we're currently ignoring physics?
		int8_t scroll_x = -half_rotation; 		// Much cleaner. 
		if( rotation > 64 ) { scroll_x = -64 + half_rotation; } 		// Past the horizon, reverse back toward 0. 
		if( rotation < -64 ) { scroll_x = 64 + half_rotation; } 

		int8_t scroll_y = ( abs( half_rotation ) >> 2 );  		// Don't reverse Y when upside-down. 
		// Nudge screen upward when crossing panel edges? Prior notes:
			// Make xstart follow rotation, so the camera tracks a car going up the wall. 
				// I'm okay with the chunky motion up and down, because it kinda matches where your car hits an angled wall. 
			// (Though sloppy correlation is fine because I do like the coincidental transition between facets.) 


		// Debug - no sliding around while dorking with curves:
		scroll_x = scroll_y = 0; 
		scroll_x += pan_x; 

		// The simple fix is to always render 32x18. That'll at least let us figure out if we're tracking sensibly. Perf later. 
		// Setting SCX/SCY_REG eventually has to go in the vblank interrupt, for when we do double-buffering and vsync. 
		uint8_t scx = scroll_x + 8*((32-20)/2); 		// Okay, sure. 8x for bytes, not bits. 
		uint8_t scy = scroll_y + 8*(32-18)/2 + 32; 		// Jumble of constants is to center the screen. (And offset Y.) 
		SCX_REG = scx; 		// This indirection (scx -> SCX_REG) might be redundant. It was intended for ystart etc. 
		SCY_REG = scy; 

		// Rendering 
		// Simplify panning: the logical screen is always 32x32 tiles. We just try not to draw what we can't see. 
		// The stupid version of this is to check x and y in each loop and "continue" if they're fully offscreen. 


		// Displacement curves don't work. 
		// I'm gonna try reordering how we draw tiles, so we can pretend we're drawing rectangles, 
			// e.g. 32x32, then 31x31, and so on, with the center point moving between each ring of tiles. 
		// Niavely: for ring = 32..0. Choose offset in whole tiles. (Might be end_offset / 32. Might be offset += per_ring.)
			// for y=ring, x=(32-ring)..ring, draw a row.
			// for y=ring+1..(32-(ring+1)), draw x=(32-ring) and x=ring. 
			// for y=(32-ring), x=(32-ring..ring), draw a row. 

//		jitter = scroller_x; 
		// Note: for-y and for-y loops don't use WIDTH / HEIGHT because we're still only drawing 20x18. 
		// "Evelyn" keeps barking about this: 
		const int8_t ystart = scy / 8; 
		const int8_t xstart = scx / 8; 
		const int8_t ystop = ystart + 17; 
		const int8_t xstop = xstart + 21;  
			// Screen is 18 tiles tall. +1 for overlap... except we're kinda covering up the bottom with some HUD stuff. 
			// Yeah, place the window to cover up two rows. We get a whole extra frame per second. 

/*
		uint8_t row = ystart;  
		for( int8_t y = ystart; y < ystop; y++ ) { 
			vram_pointer = get_bkg_xy_addr( xstart, row++ );
			for( int8_t x = xstart; x < xstop; x++ ) {  
*/
		// Alternate for-loop head, for displacement curves:
		bend_x = 0; 
		bend_y = 0;  
			// Past 7, you get very sudden inclusion of overflow. (May still be reversal past 5.) 
		xplus = 1; 		// Every tile on top / bottom of ring, skip from first to last elsewhere. 
		for( int8_t ring = 8; ring > 0; ring-- ) { 		// Oh, this only needs to go from 32..16, since they meet in the middle. 
	//		uint8_t ring=0; {		// Debug 
			bend_x += curve_x; 		// Should probably be more like scroll_x / 32. 
			bend_y += curve_y; 

			// Well, this doesn't work right, but it's wrong fast. 
			// Am I not setting x and y correctly? That shouldn't be it - it'd draw garbage, with vram_pointer++. 
			// And anyway, these are ortho. Top/bottom bar, y is constant, side brackets, x is constant. 
			// Oh, for-- the side brackts have a vram_pointer assignment inside the for-loop.
			// Top's still fucked up, though. Durr, x+=xplus. 

			// Top bar
			// This alone should draw a triangle from top edge to center. 
			// vram_pointer gets xplus added in every draw_tile call. 
			y = 16 - ring; 
			xplus = 1; 		// Next tile over
			vram_pointer = get_bkg_xy_addr( 16 - ring, y );
			for( x = 16 - ring; x < 16 + ring; x += xplus ) { draw_tile(); } 

			// Left bracket
			xplus = 32; 		// Next tile down
			x = 16 - ring; 
			vram_pointer = get_bkg_xy_addr( x, 16 - ring + 1 );
			for( y = 16 - ring + 1; y < 16 + ring - 1; y++ ) { 
//				vram_pointer = get_bkg_xy_addr( 16 - ring, y );
//				for( x = 16 - ring; x < 16 + ring; x += 2*ring-1 ) { draw_tile(); } 
				draw_tile();
			}

			// Right bracket
			xplus = 32; 		// Next tile down 
			x = 16 + ring - 1; 
			vram_pointer = get_bkg_xy_addr( x, 16 - ring + 1 );
			for( y = 16 - ring + 1; y < 16 + ring - 1; y++ ) { 
//				vram_pointer = get_bkg_xy_addr( 16 - ring, y );
//				for( x = 16 - ring; x < 16 + ring; x += 2*ring-1 ) { draw_tile(); } 
				draw_tile();
			}

			// Bottom bar 
			y = 16 + ring - 1; 
			xplus = 1; 		// Next tile over
			vram_pointer = get_bkg_xy_addr( 16 - ring, y );
			for( x = 16 - ring; x < 16 + ring; x += xplus ) { draw_tile(); } 


#if 0
//			vram_pointer = get_bkg_xy_addr( 16 - ring, 16 - ring );
			for( y = 16-ring; y < 16+ring; y++ ) { 		// ring counts down. ring is the higher value. 
	//			xplus = 2*ring - 1; 		// E.g. ring=8 for 16x16 -> xplus is -8, then... +7? 
					// Why the fuck is this so broken? No values make any goddamn sense. 
					// It should be ring-1, so x=(32-ring)+ring-1 oh I see. 
					// So should it be ring=16..0? And then xplus is... fuck coordinate systems. 
					// For starters ring should be going down by 2 if it's 32..0. 
					// Went for 16..0. And it's doing stuff between? What the fuck? 
	//			if( y == (32-ring) || y == ring-1 ) { xplus = 1; } 
//				if( y == (16-ring) || y == 15+ring ) { xplus = 1; } else { xplus = 2*ring-1; } 
//				vram_pointer = get_bkg_xy_addr( 16 - ring, y );
//				vram_pointer += 32 - xplus; 

		 		// Turns out functions are relatively cheap. 
				// draw_tile() implicitly takes x, y, xplus, bend_x, bend_y, and vram_pointer.
				// Hang on, no. This goes outside of for-y. 
				vram_pointer = get_bkg_xy_addr( 16 - ring, y );
				for( x = 16 - ring; x < 16 + ring; x += 2*ring-1 ) { draw_tile(); } 

				vram_pointer = get_bkg_xy_addr( 16 - ring, y );
				for( x = 16 - ring; x < 16 + ring; x += 2*ring-1 ) { draw_tile(); } 

				vram_pointer = get_bkg_xy_addr( 16 - ring, y );
				for( x = 16 - ring; x < 16 + ring; x += 2*ring-1 ) { draw_tile(); } 
			} 		// for y
#endif

		} 		// for ring

		// Print framerate. Fine performance can be tracked by updating during hblank routine, if needed / applicable. 
		// printf is useless because you can't point it at the right place. 
		// Numeral sprites start at 16. 
		// Oh of course these have a meaningful impact on framerate. Possibly just because of division. 
		const uint16_t this_frame = line_count; 
		line_count = 0; 
		running_average = ( running_average << 3 ) - running_average + this_frame; 
		running_average = running_average >> 3; 

		const uint16_t fps = ( 60 * 144 ) / ( running_average / 10 ); 

		vram_pointer = get_win_xy_addr( 0, 0 );
		set_vram_byte( ( vram_pointer++ ),  ( (fps / 100) % 10 ) + 16 );
		set_vram_byte( ( vram_pointer++ ),  ( (fps / 10) % 10 ) + 16 );
 		set_vram_byte( ( vram_pointer++ ),  14 ); 		// Decimal point. 
		set_vram_byte( ( vram_pointer++ ),  ( (fps ) % 10 ) + 16 );

		// Display raw line_count as well, because why not. 
		// (Or whatever else I want to track.)
		vram_pointer = get_win_xy_addr( 6, 0 );
		set_vram_byte( ( vram_pointer++ ),  ( (curve_x / 100) % 10 ) + 16 );
		set_vram_byte( ( vram_pointer++ ),  ( (curve_x / 10) % 10 ) + 16 );
		set_vram_byte( ( vram_pointer++ ),  ( (curve_x ) % 10 ) + 16 );

		// We don't want to wait unless we're hitting 60 Hz, which, no. 
//		wait_vbl_done();	
	}
}






void draw_tile() { 
					int8_t effective_x = x + ( bend_x >> 3 ); 
					int8_t effective_y = y;// + ( bend_x >> 3 ); 
					if( effective_x > 32 ) { effective_x = 32; } 
	//				if( effective_x < 0 ) { effective_x = 0; } 		// Hmm.

					// DRY: use variable for sampling shenanigans. (Eventually for Y as well.)
					distance = distance_table[ effective_x + 32 * effective_y ] + 2*scroller_x; 
					angle = angle_table[ effective_x + 32 * effective_y ] + rotation;// + 17; 

					// Different tube patterns: 
	//				if( angle >> 5 & 0x1 ) { outbyte ^= 0x02; }
	//				outbyte = ( angle >> 5 ); 		// Faux lighting from upper-left. Super good, super easy. 
	//				outbyte += ( ( distance - 2 ) & 0x40 ) >> 6; 		// Reasonable pipe segments. &0x20>>5 also works. 
	//				outbyte += ( angle - 64 - 8 ) >> 4; 	// Can I do dark top / light bottom, as a bidirectional gradient? Punt. 
	//				outbyte = ( angle >> 4 ); 		// 16 shades (1) obviously isn't fake polygons and (2) isn't aligned to +17. 
						// Or I guess it is exactly as aligned, but previously, we wanted the midpoint in the center of 1/8th. 

					// Get some cheap division into six slices. Doesn't need to be perfect. 
					// Use a table. Duh. 
					outbyte = tube_color[ angle ]; 
	//				outbyte += distance >> 4;  		// Ooh, that's a nice spiral. 
						// Even Manhattan-distance deformation look pretty nice with this. 

	//				outbyte += distance >> 3; 		// Debug: plain rings for clear distance. 

					// Distance rings. 
					// Soft rings, now that we have smoother shades and faux anti-aliasing. 
					// Eventually mix up splits-and-struts versus rings-and-path. 
						// Maybe also classic checkerboard. Cosmetic variety. 
					if( ( distance & 0b110000 ) == 0 ) { 		// 0b10000 is much thicker. 0b110000, so they're further apart?
	//					outbyte += 2; 
	//					if( angle > 128+21 && angle < 256 - 21  ) { 		// Hex, end short of midpoint
							outbyte -= ( distance & 0b00001111 ) >> 2; 		// Oh that's nice. 
	//					}
					}

					// Omnipresent bottom path. 
					// Diagonals? Angle + distance. Eh. Kinda works, but it feels spiral-y. 
					const uint8_t diagonal = angle - 4*43; 
					if( diagonal < 43 ) { outbyte += distance & 0b1100 ? 0 : 1; } 

					outbyte &= 0x0F; 		// 16 shades (at present)
					outbyte |= 0x80; 		// ... placed halfway up the tile list. 

					// Gonna bet set_vram_byte is very safe and therefore dog-slow. (Yep.)
					// With more going on, I'm not sure this has a meaningful impact performance. 
					set_vram_byte( vram_pointer, outbyte );
//					scroller_next_char++; 
	//				vram_pointer++; 
					vram_pointer += xplus; 

} 










